from .annotate.models import (
    AnnotatedModel,
    AnnotationType,
    Annotation,
)
from .base.models import (
    LabelModel,
)
from .file.models import (
    FiledModel,
    FileType,
    File,
)
from .identify.models import (
    IdentifiedModel,
    Authority,
    Identifier,
    IdentifiedModelManager,
)
from .locate.models import (
    LocatedModel,
    Location,
)
from .mark.models import (
    MarkedModel,
    MarkdownType,
    Markdown,
)
from .relate.models import (
    RelatedModel,
    RelationshipType,
    Relationship,
    InverseRelationshipType,
    InverseRelationship,
    RelatedModelManager,
)

__all__ = (
    'AnnotatedModel',
    'AnnotationType',
    'Annotation',

    'LabelModel',

    'FiledModel',
    'FileType',
    'File',

    'IdentifiedModel',
    'Authority',
    'Identifier',

    'LocatedModel',
    'Location',

    'MarkedModel',
    'MarkdownType',
    'Markdown',

    'RelatedModel',
    'RelationshipType',
    'Relationship',
    'InverseRelationshipType',
    'InverseRelationship',

    'IdentifiedAnnotatedModel',
    'IdentifiedLocatedModel',
    'IdentifiedMarkedModel',
    'IdentifiedRelatedModel',
)


class IdentifiedAnnotatedModel(IdentifiedModel, AnnotatedModel):
    class Meta(IdentifiedModel.Meta):
        abstract = True


class IdentifiedLocatedModel(IdentifiedModel, LocatedModel):
    class Meta(IdentifiedModel.Meta):
        abstract = True


class IdentifiedMarkedModel(IdentifiedModel, MarkedModel):
    class Meta(IdentifiedModel.Meta):
        abstract = True


class IdentifiedRelatedModelManager(
        IdentifiedModelManager, RelatedModelManager):
    pass


class IdentifiedRelatedModel(IdentifiedModel, RelatedModel):
    objects = IdentifiedRelatedModelManager()

    class Meta(IdentifiedModel.Meta):
        abstract = True



from rest_framework import serializers
from rest_framework.validators import UniqueTogetherValidator
from wq.db.patterns.base import serializers as base
from .models import Identifier


class IdentifierListSerializer(base.TypedAttachmentListSerializer):
    def to_internal_value(self, data):
        data = super(IdentifierListSerializer, self).to_internal_value(data)
        primary = [
            ident for ident in data
            if ident and ident.get('is_primary', None)
        ]
        if not any(primary) and len(data) > 0:
            for ident in data:
                if ident:
                    ident['is_primary'] = True
                    break
        return data


class IdentifierSerializer(base.TypedAttachmentSerializer):
    url = serializers.ReadOnlyField()

    class Meta(base.TypedAttachmentSerializer.Meta):
        model = Identifier
        list_serializer_class = IdentifierListSerializer

        # patterns-specific
        type_field = 'authority_id'


class IdentifiedModelValidator(UniqueTogetherValidator):
    def enforce_required_fields(self, attrs):
        pass

    def filter_queryset(self, attrs, queryset):
        for field in self.fields:
            attrs.setdefault(field, None)
        return super(IdentifiedModelValidator, self).filter_queryset(
            attrs, queryset
        )


class IdentifiedModelSerializer(base.AttachedModelSerializer):
    identifiers = IdentifierSerializer(many=True)
    slug = serializers.SlugField(required=False)

    def get_unique_together_validators(self):
        return [IdentifiedModelValidator(
            queryset=self.Meta.model.objects.all(),
            fields=['slug']
        )]

    def save(self, *args, **kwargs):
        super(IdentifiedModelSerializer, self).save(*args, **kwargs)
        # Fetch instance from DB in case identifier changed slug while saving
        self.instance = self.Meta.model.objects.get(pk=self.instance.pk)
        return self.instance

    class Meta:
        wq_config = {
            'lookup': 'slug',
        }



from rest_framework import serializers
from wq.db.patterns.base import serializers as base
from .models import File, FileField
from django.core.files.uploadedfile import UploadedFile


class FileListSerializer(base.TypedAttachmentListSerializer):
    def get_value(self, dictionary):
        values = dictionary.get(self.source, None)
        if not isinstance(values, list):
            values = [values]
        if all(isinstance(value, UploadedFile) for value in values):
            return [
                {'file': value} for value in values
            ]
        return super(FileListSerializer, self).get_value(dictionary)


class FileSerializer(base.TypedAttachmentSerializer):
    is_image = serializers.ReadOnlyField()

    def __init__(self, *args, **kwargs):
        self.serializer_field_mapping[FileField] = serializers.FileField
        super(FileSerializer, self).__init__(*args, **kwargs)

    class Meta(base.TypedAttachmentSerializer.Meta):
        list_serializer_class = FileListSerializer
        model = File
        wq_config = {
            'initial': None,
        }


class FiledModelSerializer(base.AttachedModelSerializer):
    files = FileSerializer(many=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-01-17 03:47
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import wq.db.patterns.file.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('file', wq.db.patterns.file.models.FileField(height_field='height', upload_to='.', width_field='width')),
                ('size', models.IntegerField(blank=True, null=True)),
                ('width', models.IntegerField(blank=True, null=True)),
                ('height', models.IntegerField(blank=True, null=True)),
                ('object_id', models.PositiveIntegerField(db_index=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'db_table': 'wq_file',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='FileType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('mimetype', models.CharField(max_length=255)),
            ],
            options={
                'db_table': 'wq_filetype',
            },
        ),
        migrations.AddField(
            model_name='file',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='file.FileType'),
        ),
    ]



from wq.db.patterns.base import serializers as base
from .models import Annotation


class AnnotationSerializer(base.TypedAttachmentSerializer):
    class Meta(base.TypedAttachmentSerializer.Meta):
        model = Annotation


class AnnotatedModelSerializer(base.AttachedModelSerializer):
    annotations = AnnotationSerializer(many=True)



from rest_framework import serializers
from wq.db.rest.serializers import ModelSerializer
from wq.db.patterns.base import serializers as base
from wq.db.rest.models import get_by_identifier, ContentType

from .models import (
    Relationship, InverseRelationship,
    RelationshipType, InverseRelationshipType
)


class ContentTypeField(serializers.SlugRelatedField):
    queryset = ContentType.objects.all()
    slug_field = "model"

    def __init__(self, **kwargs):
        super(serializers.RelatedField, self).__init__(**kwargs)


class RelationshipSerializer(base.TypedAttachmentSerializer):

    def get_fields(self):
        fields = super(RelationshipSerializer, self).get_fields()
        fields['type_label'].source = 'reltype'
        return fields

    def to_representation(self, rel):
        data = super(RelationshipSerializer, self).to_representation(rel)
        data.update(rel.right_dict)
        return data

    def to_internal_value(self, attachment):
        if 'item_id' in attachment and 'type_id' in attachment:
            type = self.Meta.type_model.objects.get(pk=attachment['type_id'])
            cls = type.right.model_class()
            obj = get_by_identifier(cls.objects, attachment['item_id'])
            attachment[self.Meta.item_id_field] = obj.pk
            attachment[self.Meta.item_ct_field] = type.right.pk
            del attachment['item_id']
        return super(RelationshipSerializer, self).to_internal_value(
            attachment
        )

    class Meta(base.TypedAttachmentSerializer.Meta):
        model = Relationship

        # patterns-specific
        object_field = 'left'

        # relate-specific
        type_model = RelationshipType
        item_id_field = 'to_object_id'
        item_ct_field = 'to_content_type_id'

        exclude = (
            'from_content_type', 'from_object_id',
        )


class InverseRelationshipSerializer(RelationshipSerializer):
    class Meta(RelationshipSerializer.Meta):
        model = InverseRelationship

        type_model = InverseRelationshipType
        item_id_field = 'from_object_id'
        item_ct_field = 'from_content_type_id'

        exclude = (
            'to_content_type', 'to_object_id'
        )


class RelationshipTypeSerializer(ModelSerializer):
    from_type = ContentTypeField()
    to_type = ContentTypeField()


class RelatedModelSerializer(base.AttachedModelSerializer):
    relationships = RelationshipSerializer(many=True)
    inverserelationships = InverseRelationshipSerializer(many=True)

    def set_parent_object(self, attachment, instance, name):
        if name == "relationships":
            attachment['from_content_object'] = instance
        elif name == "inverserelationships":
            attachment['to_content_object'] = instance
        else:
            super(RelatedModelSerializer, self).set_parent_object(
                attachment, instance, name
            )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-01-12 03:18
from __future__ import unicode_literals

from django.db import migrations, models
import swapper
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        swapper.dependency('mark', 'MarkdownType'),
    ]

    operations = [
        migrations.CreateModel(
            name='MarkdownType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting('mark', 'MarkdownType'),
                'db_table': 'wq_markdowntype',
            },
        ),
        migrations.CreateModel(
            name='Markdown',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('summary', models.CharField(blank=True, max_length=255, null=True)),
                ('markdown', models.TextField(blank=True, null=True)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=swapper.get_model_name('mark', 'MarkdownType'))),
            ],
            options={
                'abstract': False,
                'db_table': 'wq_markdown',
            },
        ),
    ]



from .base.serializers import (
    AttachmentSerializer,
    TypedAttachmentSerializer,
    AttachedModelSerializer,
    NaturalKeyModelSerializer,
    NaturalKeyAttachedModelSerializer,
)

from .annotate.serializers import AnnotatedModelSerializer
from .file.serializers import FiledModelSerializer
from .identify.serializers import IdentifiedModelSerializer
from .locate.serializers import LocatedModelSerializer
from .mark.serializers import MarkedModelSerializer
from .relate.serializers import RelatedModelSerializer


__all__ = (
    'AttachmentSerializer',
    'TypedAttachmentSerializer',
    'AttachedModelSerializer',
    'NaturalKeyModelSerializer',
    'NaturalKeyAttachedModelSerializer',

    'AnnotatedModelSerializer',
    'FiledModelSerializer',
    'IdentifiedModelSerializer',
    'LocatedModelSerializer',
    'MarkedModelSerializer',
    'RelatedModelSerializer',

    'IdentifiedAnnotatedModelSerializer',
    'IdentifiedLocatedModelSerializer',
    'IdentifiedMarkedModelSerializer',
    'IdentifiedRelatedModelSerializer',
)


class IdentifiedAnnotatedModelSerializer(
        IdentifiedModelSerializer, AnnotatedModelSerializer):
    pass


class IdentifiedLocatedModelSerializer(
        IdentifiedModelSerializer, LocatedModelSerializer):
    class Meta:
        wq_config = {
            'map': True,
            'lookup': 'slug',
        }


class IdentifiedMarkedModelSerializer(
        IdentifiedModelSerializer, MarkedModelSerializer):
    pass


class IdentifiedRelatedModelSerializer(
        IdentifiedModelSerializer, RelatedModelSerializer):
    pass



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import (
    GenericForeignKey, GenericRelation
)
from ..base.models import LabelModel
from django import forms


from django.conf import settings
INSTALLED = 'wq.db.patterns.file' in settings.INSTALLED_APPS


# Custom FileField handles both images and files
class FileField(models.ImageField):
    # Use base forms.FileField to skip ImageField validation
    def formfield(self, **kwargs):
        kwargs['form_class'] = forms.FileField
        return super(FileField, self).formfield(**kwargs)

    # Only update_dimension_fields for images
    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        if (getattr(instance, 'mimetype', None) is not None and
                'image' in instance.mimetype):
            super(FileField, self).update_dimension_fields(instance, force,
                                                           *args, **kwargs)
        else:
            pass

    # Allow model to specify upload directory
    def generate_filename(self, instance, filename):
        if hasattr(instance, 'get_directory'):
            self.upload_to = instance.get_directory()
        return super(FileField, self).generate_filename(instance, filename)


class FileType(LabelModel):
    name = models.CharField(max_length=255)
    mimetype = models.CharField(max_length=255)

    wq_label_template = "{{name}} ({{mimetype}})"

    class Meta:
        db_table = 'wq_filetype'
        abstract = not INSTALLED


class FileManager(models.Manager):
    def get_queryset(self):
        qs = super(FileManager, self).get_query_set()
        if self.model.type_name is not None:
            return qs.filter(type__name=self.model.type_name)
        else:
            return qs


class File(models.Model):
    name = models.CharField(max_length=255, null=True, blank=True)
    type = models.ForeignKey(FileType, null=True, blank=True)
    file = FileField(upload_to='.', width_field='width', height_field='height')
    size = models.IntegerField(null=True, blank=True)
    width = models.IntegerField(null=True, blank=True)
    height = models.IntegerField(null=True, blank=True)

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = GenericForeignKey()

    type_name = "File"

    def get_directory(self):
        if self.is_image:
            return 'images'
        else:
            return 'files'

    @property
    def mimetype(self):
        if self.type is not None:
            return self.type.mimetype

        if self.file.name not in (None, ""):
            from wq.io.util import guess_type
            mimetype = guess_type(self.file.name, self.file.read(1024))
            return mimetype
        else:
            return None

    @property
    def is_image(self):
        return self.mimetype is not None and self.mimetype.startswith('image/')

    def save(self, *args, **kwargs):
        if self.type is None:
            self.type, isnew = FileType.objects.get_or_create(
                mimetype=self.mimetype,
                name=self.type_name
            )
        if self.size is None:
            self.size = self.file.size
        if self.name is None or self.name == '':
            self.name = self.file.name
        super(File, self).save(*args, **kwargs)

    def __str__(self):
        if self.name is not None:
            return self.name
        else:
            return 'File %s' % self.id

    class Meta:
        db_table = 'wq_file'
        ordering = ("name",)
        abstract = not INSTALLED


class FiledModel(models.Model):
    files = GenericRelation(File)

    class Meta:
        abstract = True



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import (
    GenericForeignKey, GenericRelation
)
from ..base.models import LabelModel
from django.utils.translation import get_language_from_request
from django.conf import settings

import swapper
swapper.set_app_prefix('mark', 'WQ')

INSTALLED = ('wq.db.patterns.mark' in settings.INSTALLED_APPS)


class BaseMarkdownType(LabelModel):
    name = models.CharField(max_length=100)

    @classmethod
    def get_current(cls, request=None):
        if request:
            try:
                kwargs = cls.get_current_filter(request)
                return cls.objects.get(**kwargs)
            except cls.DoesNotExist:
                pass
        return cls.get_default()

    @classmethod
    def get_current_filter(cls, request):
        raise NotImplementedError()

    @classmethod
    def get_default(cls):
        markdowns = cls.objects.all()
        if len(markdowns) > 0:
            return markdowns[0]

    class Meta:
        abstract = True
        ordering = ['pk']


class MarkdownType(BaseMarkdownType):
    @classmethod
    def get_current_filter(cls, request):
        lang = get_language_from_request(request)
        return {'name': lang}

    class Meta:
        swappable = swapper.swappable_setting('mark', 'MarkdownType')
        db_table = "wq_markdowntype"
        abstract = not INSTALLED


class Markdown(models.Model):
    type = models.ForeignKey(swapper.get_model_name('mark', 'MarkdownType'))
    summary = models.CharField(max_length=255, null=True, blank=True)
    markdown = models.TextField(null=True, blank=True)

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()

    def __str__(self):
        return self.summary or ''

    @property
    def html(self):
        from markdown import markdown
        extensions = getattr(settings, 'MARKDOWN_EXTENSIONS', [])
        return markdown(self.markdown, extensions)

    class Meta:
        db_table = "wq_markdown"
        abstract = not INSTALLED


class MarkedModel(models.Model):
    markdown = GenericRelation(Markdown)

    def get_markdown(self, type):
        markdowns = self.markdown.filter(type=type)
        if len(markdowns) == 0:
            markdowns = self.markdown.order_by("type")
        if len(markdowns) > 0:
            return markdowns[0]
        return None

    def get_html(self, type):
        markdown = self.get_markdown(type)
        if markdown:
            return markdown.html
        return None

    class Meta:
        abstract = True



from rest_framework.filters import BaseFilterBackend

from wq.db.rest.models import get_ct, get_by_identifier
from .models import get_related_parents


class RelatedFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        ctype = get_ct(view.model)
        filter = {}
        for key, val in list(view.kwargs.items()) + list(request.GET.items()):
            if not key.startswith('related_'):
                continue
            if isinstance(val, list):
                val = val[0]
            for pct in get_related_parents(ctype):
                if key == 'related_' + pct.identifier:
                    pclass = pct.model_class()
                    parent = get_by_identifier(pclass.objects, val)
                    objs = view.model.objects.filter_by_related(parent)
                    filter['pk__in'] = objs.values_list('pk', flat=True)
        return queryset.filter(**filter)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import (
    GenericForeignKey, GenericRelation
)
from natural_keys import NaturalKeyModel
from ..base.models import LabelModel
from collections import OrderedDict

from django.conf import settings
INSTALLED = ('wq.db.patterns.annotate' in settings.INSTALLED_APPS)


class AnnotationType(NaturalKeyModel, LabelModel):
    name = models.CharField(max_length=255)

    class Meta:
        unique_together = [['name']]
        db_table = 'wq_annotationtype'
        abstract = not INSTALLED


class AnnotationManager(models.Manager):

    # Collapse annotations into dict for simple access
    @property
    def vals(self):
        if (hasattr(self, '_vals')):
            return getattr(self, '_vals')

        vals = OrderedDict()
        for annot in self.all():
            vals[annot.type.natural_key()[0]] = annot.value

        setattr(self, '_vals', vals)
        return vals

    # Default implementation of get_or_create doesn't work well with generics
    def get_or_create(self, **kwargs):
        try:
            return self.get(**kwargs), False
        except self.model.DoesNotExist:
            return self.create(**kwargs), True


class Annotation(models.Model):
    type = models.ForeignKey(AnnotationType)
    value = models.TextField(null=True, blank=True)

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = GenericForeignKey()

    objects = AnnotationManager()

    def __str__(self):
        return (
            '%(object)s -> %(annot)s: %(value)s'
            % {
                'object': self.content_object,
                'annot': self.type,
                'value': self.value
            })

    class Meta:
        db_table = 'wq_annotation'
        abstract = not INSTALLED


class AnnotatedModel(models.Model):
    annotations = GenericRelation(Annotation)

    @property
    def vals(self):
        return self.annotations.vals

    @vals.setter
    def vals(self, vals):
        keys = [(key,) for key in vals.keys()]
        types, success = AnnotationType.objects.resolve_keys(keys)
        if not success:
            missing = [name for name, atype in types.items() if atype is None]
            raise TypeError(
                "Could not identify one or more annotation types: %s!"
                % missing
            )

        for name, atype in types.items():
            annot, is_new = self.annotations.get_or_create(type=atype)
            annot.value = vals[name[0]]
            annot.save()

    class Meta:
        abstract = True



from rest_framework import serializers
from wq.db.patterns.base import serializers as base
import swapper

from .models import Markdown
MarkdownType = swapper.load_model(
    'mark', 'MarkdownType', required=False
)


def active_markdown(qs, request):
    mtype = MarkdownType.get_current(request)
    return qs.filter(type=mtype)


class MarkdownSerializer(base.TypedAttachmentSerializer):
    html = serializers.ReadOnlyField()

    class Meta(base.TypedAttachmentSerializer.Meta):
        model = Markdown
        list_exclude = ('html',)


class MarkedModelSerializer(base.AttachedModelSerializer):
    markdown = MarkdownSerializer.for_depth(1)(many=True)

    def to_representation(self, instance):
        data = super(MarkedModelSerializer, self).to_representation(instance)
        if 'markdown' not in data or 'request' not in self.context:
            return data

        # Only include active markdown in output
        # (FIXME: ideally this filter would happen *before* initial
        #  serialization)
        request = self.context['request']
        active_ids = active_markdown(
            instance.markdown, request
        ).values_list('pk', flat=True)
        data['markdown'] = [
            markdown for markdown in data['markdown']
            if markdown['id'] in active_ids
        ]
        return data



from rest_framework.generics import GenericAPIView as RestGenericAPIView
from rest_framework.response import Response
from rest_framework.decorators import detail_route
from rest_framework import status, viewsets
from .model_tools import get_ct, get_object_id, get_by_identifier
from django.db.models.fields import FieldDoesNotExist


class GenericAPIView(RestGenericAPIView):
    router = None
    ignore_kwargs = []

    @property
    def template_name(self):
        """
        Infer template name from view/viewset name
        """
        name = type(self).__name__
        name = name.replace('ViewSet', '')
        name = name.replace('View', '')
        return name.lower() + '.html'

    @property
    def depth(self):
        return 0

    def get_template_names(self):
        return [self.template_name]

    def get_queryset(self):
        if self.router is not None and self.model is not None:
            return self.router.get_queryset_for_model(self.model, self.request)
        return super(GenericAPIView, self).get_queryset()

    def get_serializer_class(self):
        if self.router is not None and self.model is not None:
            return self.router.get_serializer_for_model(self.model, self.depth)
        return super(GenericAPIView, self).get_serializer_class()


class SimpleView(GenericAPIView):
    def get(self, request, *args, **kwargs):
        return Response({})


class SimpleViewSet(viewsets.ViewSet, GenericAPIView):
    def list(self, request, *args, **kwargs):
        return Response({})


class ModelViewSet(viewsets.ModelViewSet, GenericAPIView):
    target = None

    @property
    def template_name(self):
        basename = get_ct(self.model).identifier
        if self.action in ('retrieve', 'create', 'update', 'delete'):
            suffix = 'detail'
        elif self.action and self.action != 'metadata':
            suffix = self.action
        else:
            suffix = 'list'
        return "%s_%s.html" % (basename, suffix)

    @property
    def depth(self):
        if self.action in ('retrieve', 'edit'):
            return 1
        else:
            return 0

    @detail_route()
    def edit(self, request, *args, **kwargs):
        """
        Generates a context appropriate for editing a form
        """
        response = self.retrieve(request, *args, **kwargs)
        self.add_lookups(response.data)
        return response

    def new(self, request):
        """
        new is a variant of the "edit" action, but with no existing model
        to lookup.
        """
        self.action = 'edit'
        init = request.GET.dict()
        for arg in self.ignore_kwargs:
            init.pop(arg, None)
        for key in list(init.keys()):
            try:
                field = self.model._meta.get_field(key)
            except FieldDoesNotExist:
                del init[key]
            else:
                if field.rel:
                    fk_model = field.rel.to
                    try:
                        obj = get_by_identifier(fk_model.objects, init[key])
                    except fk_model.DoesNotExist:
                        del init[key]
                    else:
                        init[key] = obj.pk

        obj = self.model(**init)
        serializer = self.get_serializer(obj)
        data = serializer.data
        self.add_lookups(data)
        return Response(data)

    def retrieve(self, request, *args, **kwargs):
        """
        Custom retrieve watches for "new" lookup value and switches modes
        accordingly
        """
        if hasattr(self, 'lookup_url_kwarg'):
            # New in DRF 2.4?
            lookup = self.lookup_url_kwarg or self.lookup_field
        else:
            lookup = self.lookup_field

        if self.kwargs.get(lookup, "") == "new":
            # new/edit mode
            return self.new(request)
        else:
            # Normal detail view
            return super(ModelViewSet, self).retrieve(request, *args, **kwargs)

    def add_lookups(self, context):
        # Mimic _addLookups in wq.app/app.js
        context['edit'] = True

        if not self.router:
            return

        ct = get_ct(self.model)
        conf = ct.get_config()

        for field in conf['form']:
            if 'choices' in field:
                # CharField choices
                context[field['name'] + '_choices'] = [{
                    'name': choice['name'],
                    'label': choice['label'],
                    'selected': (
                        choice['name'] == context.get(field['name'], '')
                    ),
                } for choice in field['choices']]

            elif 'wq:ForeignKey' in field:
                choices = self.get_lookup_choices(field, context)
                if choices:
                    context[field['name'] + '_list'] = choices

    def get_lookup_choices(self, field, context):
        model_name = field['wq:ForeignKey']
        model_conf = self.router.get_model_config(field['wq:ForeignKey'])
        if not model_conf:
            return

        qs = self.router.get_queryset_for_model(model_name, self.request)
        if field.get('filter', None):
            qs = qs.filter(**self.compute_filter(
                field['filter'],
                model_conf,
                context
            ))
        choices = self.serialize_choices(qs, field)
        self.set_selected(choices, context.get(field['name'] + '_id', ''))
        return choices

    def compute_filter(self, filter, model_conf, context):
        def render(value):
            import pystache
            result = pystache.render(value, context)
            if result.isdigit():
                result = int(result)
            return result

        fk_lookups = {}
        for field in model_conf['form']:
            if 'wq:ForeignKey' not in field:
                continue
            lookup = self.router.get_lookup_for_model(
                field['wq:ForeignKey']
            )
            if lookup and lookup != 'pk':
                fk_lookups['%s_id' % field['name']] = '%s__%s' % (
                    field['name'], lookup
                )

        computed_filter = {}
        for key, values in filter.items():
            if not isinstance(values, list):
                values = [values]
            values = [
                render(value) if '{{' in value else value
                for value in values
            ]

            if key in fk_lookups:
                key = fk_lookups[key]

            if len(values) > 1:
                computed_filter[key + '__in'] = values
            else:
                computed_filter[key] = values[0]

        return computed_filter

    def serialize_choices(self, qs, field):
        return [{
            'id': get_object_id(obj),
            'label': str(obj)
        } for obj in qs]

    def set_selected(self, choices, value):
        for choice in choices:
            if choice['id'] == value:
                choice['selected'] = True

    def list(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).list(
            request, *args, **kwargs
        )
        if not isinstance(response.data, dict):
            return response

        if self.target:
            response.data['target'] = self.target
        ct = get_ct(self.model)
        for pct, fields in ct.get_foreign_keys().items():
            if len(fields) == 1:
                self.get_parent(pct, fields[0], response)
        return response

    def create(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).create(
            request, *args, **kwargs
        )
        if not request.accepted_media_type.startswith('text/html'):
            # JSON request, assume client will handle redirect
            return response

        # HTML request, probably a form post from an older browser
        if response.status_code == status.HTTP_201_CREATED:
            return self.postsave(request, response)
        else:
            return self.saveerror(request, response)

    def update(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).update(
            request, *args, **kwargs
        )
        if not request.accepted_media_type.startswith('text/html'):
            # JSON request, assume client will handle redirect
            return response

        # HTML request, probably a form post from an older browser
        if response.status_code == status.HTTP_200_OK:
            return self.postsave(request, response)
        else:
            return self.saveerror(request, response)

    def postsave(self, request, response):
        ct = get_ct(self.model)
        conf = ct.get_config(request.user)

        # Redirect to new page
        postsave = conf.get('postsave', ct.identifier + '_detail')
        if '_' in postsave:
            page, mode = postsave.split('_')
        else:
            page = postsave
            mode = 'detail'

        oid = ""
        if page != ct.identifier and self.router:
            # Optional: return to detail view of a parent model
            pconf = self.router.get_page_config(page)
            if pconf.get('list', None) and mode != "list":
                oid = response.data.get(page + '_id', None)
        else:
            # Default: return to detail view of the saved model
            pconf = conf
            if mode != "list":
                oid = response.data['id']

        url = "/" + pconf['url']
        if pconf['url'] and pconf.get('list', None):
            url += "/"
        if oid:
            url += str(oid)
            if mode == "edit":
                url += "/edit"

        return Response(
            {'detail': 'Created'},
            status=status.HTTP_302_FOUND,
            headers={'Location': url}
        )

    def saverror(self, request, response):
        errors = [{
            'field': key,
            'errors': val
        } for key, val in response.data.items()]
        template = get_ct(self.model).identifier + '_error.html'
        return Response(
            {
                'errors': errors,
                'post': request.DATA
            },
            status=response.status_code,
            template_name=template
        )

    def get_parent(self, ct, kwarg_name, response):
        pid = self.kwargs.get(kwarg_name, None)
        if not pid:
            return

        pcls = ct.model_class()
        if self.router:
            slug = self.router.get_lookup_for_model(pcls)
            parent = pcls.objects.get(**{slug: pid})
        else:
            parent = get_by_identifier(pcls.objects, pid)
        if ct.urlbase == '':
            urlbase = ''
        else:
            urlbase = ct.urlbase + '/'
        objid = get_object_id(parent)
        response.data['parent_label'] = str(parent)
        response.data['parent_id'] = objid
        response.data['parent_url'] = '%s%s' % (urlbase, objid)
        response.data['parent_is_' + ct.identifier] = True
        response.data['parent_page'] = ct.identifier
        response.data['page_config'] = get_ct(self.model).get_config()
        if self.router:
            response.data['parent'] = self.router.serialize(parent)
        return parent

    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)



from django.http import Http404
from rest_framework.generics import GenericAPIView
from rest_framework.mixins import ListModelMixin
from rest_framework.response import Response
from rest_framework import status
from wq.db.rest.views import SimpleView

from .util import search
from .serializers import SearchResultSerializer

SEARCH_PARAMETER = 'q'


class SearchView(SimpleView, GenericAPIView, ListModelMixin):
    auto = False
    search = None
    content_type = None
    authority_id = None
    serializer_class = SearchResultSerializer

    def get_queryset(self):
        if not self.search:
            return []
        return search(
            self.search,
            self.auto,
            self.content_type,
            self.authority_id
        )

    def filter_queryset(self, queryset):
        return queryset

    def get(self, request, *args, **kwargs):
        self.search = request.GET.get(SEARCH_PARAMETER, None)
        self.auto = request.GET.get('auto', self.auto)
        self.content_type = request.GET.get('type', self.content_type)
        self.authority_id = request.GET.get('authority_id', self.authority_id)
        response = self.list(request, args, kwargs)

        if response.data['count'] == 1 and self.auto:
            return self.auto_redirect(response)
        else:
            return response

    def auto_redirect(self, response):
        item = response.data['list'][0]
        return Response(
            status=status.HTTP_302_FOUND,
            data={'message': 'Found', 'search': self.search},
            headers={'Location': '/' + item['url']}
        )


class DisambiguateView(SearchView):
    def get(self, request, *args, **kwargs):
        self.search = kwargs['slug']
        response = self.list(request, *args, **kwargs)
        if response.data['count'] == 0:
            raise Http404("Could not find %s", self.search)
        elif response.data['count'] == 1:
            return self.auto_redirect(response)
        else:
            response.data['message'] = "Multiple Matches Found"
            return response



from rest_framework.routers import Route
from wq.db.rest.views import ModelViewSet
from wq.db.rest.models import get_ct
from .models import get_related_parents, get_related_children
from .filters import RelatedFilterBackend


class RelatedModelViewSet(ModelViewSet):
    filter_backends = ModelViewSet.filter_backends + [RelatedFilterBackend]

    def list(self, request, *args, **kwargs):
        response = super(RelatedModelViewSet, self).list(
            request, *args, **kwargs
        )
        ct = get_ct(self.model)
        for pct in get_related_parents(ct):
            self.get_parent(pct, 'related_%s' % pct.identifier, response)
        return response

    @classmethod
    def extra_routes(cls):
        routes = []
        ct = get_ct(cls.model)
        for pct in get_related_parents(ct):
            if not pct.is_registered():
                continue
            if pct.urlbase == '':
                purlbase = ''
            else:
                purlbase = pct.urlbase + '/'

            routes.append(Route(
                (
                    '^' + purlbase + r'(?P<related_' + pct.identifier +
                    '>[^\/\?]+)/{prefix}{trailing_slash}$'
                ),
                mapping={'get': 'list'},
                name="{basename}-for-related-%s" % pct.identifier,
                initkwargs={'suffix': 'List'},
            ))

        for cct in get_related_children(ct):
            if not cct.is_registered():
                continue
            cbase = cct.urlbase
            routes.append(Route(
                url='^%s-by-{prefix}' % cbase,
                mapping={'get': 'list'},
                name="%s-by-%s" % (cct.identifier, ct.identifier),
                initkwargs={'target': cbase, 'suffix': 'List'},
            ))

        return routes



from .serializers import (
    AnnotatedModelSerializer,
    FiledModelSerializer,
    IdentifiedModelSerializer,
    LocatedModelSerializer,
    MarkedModelSerializer,
    RelatedModelSerializer,

    IdentifiedAnnotatedModelSerializer,
    IdentifiedLocatedModelSerializer,
    IdentifiedMarkedModelSerializer,
    IdentifiedRelatedModelSerializer,
)


__all__ = (
    'AnnotatedModelSerializer',
    'FiledModelSerializer',
    'IdentifiedModelSerializer',
    'LocatedModelSerializer',
    'MarkedModelSerializer',
    'RelatedModelSerializer',

    'IdentifiedAnnotatedModelSerializer',
    'IdentifiedLocatedModelSerializer',
    'IdentifiedMarkedModelSerializer',
    'IdentifiedRelatedModelSerializer',
)



from rest_framework.pagination import PageNumberPagination
from collections import OrderedDict
from rest_framework.response import Response


class Pagination(PageNumberPagination):
    page_size = 50
    page_size_query_param = 'limit'

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            # DRF default metadata
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),

            # wq.db additional metadata
            ('page', self.page.number),
            ('pages', self.page.paginator.num_pages),
            ('per_page', self.page.paginator.per_page),
            ('multiple', self.page.paginator.num_pages > 1),

            # Actual data ('results' in DRF)
            ('list', data)
        ]))



from wq.db.rest.serializers import ModelSerializer
from django.conf import settings


HAS_SOCIAL_AUTH = ('social.apps.django_app.default' in settings.INSTALLED_APPS)
if HAS_SOCIAL_AUTH:
    from social.apps.django_app.default.models import UserSocialAuth

    class SocialAuthSerializer(ModelSerializer):
        def to_representation(self, instance):
            return {
                'provider_id': str(instance.provider),
                'provider_label': instance.provider.title(),
                'id': instance.pk,
                'label': instance.uid,
                'can_disconnect': type(instance).allowed_to_disconnect(
                    user=instance.user,
                    backend_name=instance.provider,
                    association_id=instance.pk
                )
            }

        class Meta:
            model = UserSocialAuth


class UserSerializer(ModelSerializer):
    if HAS_SOCIAL_AUTH:
        social_auth = SocialAuthSerializer(many=True, read_only=True)

    def to_representation(self, instance):
        result = super(UserSerializer, self).to_representation(instance)
        if (HAS_SOCIAL_AUTH and 'social_auth' in result):
            if len(result['social_auth']) > 0:
                result['social_auth'] = {'accounts': result['social_auth']}
            else:
                result['social_auth'] = None
        return result

    class Meta:
        exclude = ('id', 'password', 'user_permissions', 'groups',)
        readonly_fields = ('last_login', 'date_joined')
        list_exclude = ('social_auth',)



from wq.db.rest.models import get_ct, get_object_id
from wq.db.patterns.base import serializers as base
from django.utils.six import string_types
import json
from .models import Location


class LocationListSerializer(base.AttachmentListSerializer):
    def to_representation(self, data):
        data = super(LocationListSerializer, self).to_representation(data)
        if not self.child.is_geojson:
            return data

        if self.child.as_geometry:
            if len(data) == 1:
                return data[0]
            elif len(data) > 1:
                return {
                    'type': 'GeometryCollection',
                    'geometries': data
                }
            else:
                return None
        else:
            features = []
            for loc in data:
                feat = {
                    'id': loc['id'],
                    'type': 'Feature',
                    'geometry': loc['geometry'],
                    'properties': loc
                }
                del loc['geometry']
                del loc['id']
                features.append(feat)
            return features

    def get_value(self, dictionary):
        vals = dictionary.get(self.field_name, None)
        if isinstance(vals, string_types) and vals.strip() != '':
            vals = json.loads(vals)
        if (isinstance(vals, dict) and
                vals.get('type', None) == "FeatureCollection"):
            if 'crs' in vals:
                features = []
                for feature in vals['features']:
                    feature.update({
                        'crs': vals['crs']
                    })
                    features.append(feature)
            else:
                features = vals['features']
        else:
            features = []

        return features


class LocationSerializer(base.AttachmentSerializer):
    def __init__(self, as_geometry=False, **kwargs):
        self.as_geometry = as_geometry
        super(LocationSerializer, self).__init__(**kwargs)

    def to_representation(self, loc):
        has_parent = (
            self.parent and self.parent.parent and
            hasattr(self.parent.parent.Meta, 'model')
        )
        if self.as_geometry and has_parent:
            return json.loads(loc.geometry.geojson)

        data = super(LocationSerializer, self).to_representation(loc)
        if has_parent:
            pass
        else:
            # Include pointer to parent object (see annotate/serializers.py)
            idname = get_ct(loc.content_object).identifier + '_id'
            data[idname] = get_object_id(loc.content_object)
        return data

    def to_internal_value(self, data):
        if (data.get('type', None) == 'Feature' and
                'properties' in data and 'geometry' in data):
            obj = data['properties']
            obj['geometry'] = data['geometry']
            if 'id' in data:
                obj['id'] = data['id']
            if 'crs' in data:
                obj['geometry']['crs'] = data['crs']
            data = obj
        return super(LocationSerializer, self).to_internal_value(data)

    class Meta:
        exclude = ('content_type', 'object_id')
        model = Location
        list_serializer_class = LocationListSerializer


class LocatedModelSerializer(base.AttachedModelSerializer):
    locations = LocationSerializer(many=True)

    def get_fields(self):
        fields = super(LocatedModelSerializer, self).get_fields()
        if self.is_geojson:
            # Rename locations to "features" or "geometry" for use as GeoJSON
            locations = fields.pop('locations')
            if self.is_edit:
                name = 'features'
            else:
                name = 'geometry'
                locations.child.as_geometry = True
            locations.source = 'locations'
            fields[name] = locations
        elif not self.is_detail:
            fields.pop('locations')
        return fields

    class Meta:
        wq_config = {
            'map': True,
        }



from wq.db import rest
from .models import RelationshipType
from .serializers import RelationshipTypeSerializer


rest.router.register_model(
    RelationshipType,
    serializer=RelationshipTypeSerializer,
    fields="__all__",
)



from django.contrib.auth.models import Group
from django.conf import settings


def assign_default_group(backend, details, response, user=None,
                         is_new=False, *args, **kwargs):

    if user is None or not is_new:
        return None

    group, created = Group.objects.get_or_create(
        name=settings.DEFAULT_AUTH_GROUP
    )
    user.groups.add(group)
    return None



from rest_framework.serializers import Serializer
from wq.db.rest.models import get_ct, ContentType, get_object_id
from wq.db.patterns.models import Identifier, Annotation
from django.core.exceptions import ImproperlyConfigured


class SearchResultSerializer(Serializer):
    def to_representation(self, obj):
        if hasattr(obj, 'content_type_id') and hasattr(obj, 'content_object'):
            ctype = ContentType.objects.get(pk=obj.content_type_id)
            if isinstance(obj, Identifier):
                match_str = "%s (%s)" % (obj.name, obj.slug)
            elif isinstance(obj, Annotation):
                match_str = obj.value
            else:
                match_str = str(obj)
            obj = obj.content_object
        else:
            ctype = get_ct(obj)
            match_str = str(obj)
        if not ctype.is_registered():
            raise ImproperlyConfigured(
                "Register %s to include it in search results"
                % ctype.model_class()
            )

        obj_id = get_object_id(obj)

        url = ctype.urlbase
        if url:
            url += '/'
        url += str(obj_id)

        return {
            'id': obj_id,
            'url': url,
            'type': str(ctype),
            'label': str(obj),
            'match': match_str,
        }



from django.forms import ModelForm, Select
from itertools import chain

# Until https://code.djangoproject.com/ticket/15602 is fixed, we need a way
# to make the 'type' field readonly for existing annotations in an
# AnnotationInline. This is done here by extending the Select widget.


class SelectOnce(Select):
    def render(self, name, value, attrs=None, choices=()):
        if value is None:
            return super(SelectOnce, self).render(name, value, attrs, choices)
        else:
            for option_value, option_label in chain(self.choices, choices):
                if str(option_value) == str(value):
                    return (
                        "<p>%s</p><input type='hidden' name='%s' value='%s'>"
                        % (option_label, name, value)
                    )
        return "<p>Not Found</p>"


class AnnotationForm(ModelForm):
    class Meta:
        widgets = {
            'type': SelectOnce
        }



from wq.db import rest


def is_authenticated(request):
    return {
        'is_authenticated': request.user.is_authenticated()
    }


def social_auth(request):
    if (not request.user.is_authenticated() or
            not hasattr(request.user, 'social_auth') or
            request.user.social_auth.count() == 0):
        return {}

    user = rest.router.serialize(request.user)
    return {
        'social_auth': user.get('social_auth', {})
    }



from django.conf.urls import url
from django.conf import settings

from .views import SearchView, DisambiguateView

urlpatterns = [
    url('^search/?$', SearchView.as_view()),
    url('^search\.(?P<format>\w+)$', SearchView.as_view()),
]

if getattr(settings, 'DISAMBIGUATE', False):
    urlpatterns += [
        url('^(?P<slug>[^\/]+)$', DisambiguateView.as_view()),
    ]
